# Sherlock version 1.0.0
# SherlockSolver.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of Sherlock, a primitive solver for numerical contraint satisfaction problems.
# Sherlock is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Sherlock is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Sherlock.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Assignments
# 3) Solver
#   3.1) Generate and test
#   3.2) Find one solution
#   3.3) Find all solutions
#   3.4) Assignment management
#   3.5) Find best solution
#   3.6) Branch and bound

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from SherlockModels import *

# 2) Assignments
# --------------

class SherlockAssignment:

  def __init__(self):
    self._domains = {}
    self._parent = None

  def GetDomain(self, variable):
    domain = self._domains.get(variable, None)
    if domain==None and self._parent!=None:
      domain = self._parent.GetDomain(variable)
    return domain
  def AddDomain(self, variable, domain):
    if variable in self._domains:
      del self._domains[variable]
    self._domains[variable] = domain
  def GetDomainSize(self, variable):
    domain = self.GetDomain(variable)
    return domain.GetSize()

  def GetVariables(self):
    if self._parent!=None:
      return self._parent.GetVariables()
    return self._domains.keys()
  def GetSortedVariables(self):
    if self._parent!=None:
      return self._parent.GetSortedVariables()
    return sorted(self._domains.keys(), key=lambda var: var.GetName())

  def GetDomains(self):
    return self._domains.values()

  def GetParent(self):
    return self._parent
  def Expand(self):
    assignment = SherlockAssignment()
    assignment._parent = self
    return assignment
  def Clone(self):
    clone = SherlockAssignment()
    for variable in self._domains.keys():
      domain = self._domains[variable]
      clone._domains[variable] = domain.Clone()
    return clone

  def RemoveValue(self, variable, value):
    referenceDomain = self.GetDomain(variable)
    if not referenceDomain.DoesContainValue(value):
      return True
    domain = self._domains.get(variable, None)
    if domain==None:
      domain = referenceDomain.Clone()
      self._domains[variable] = domain
    domain.RemoveValue(value)
    # sys.stdout.write("remove value " + str(value) + " from domain of " + variable.GetName() + "\n")
    return domain.GetSize()>0
  def UpdateMinimumValue(self, variable, lowerBound):
    referenceDomain = self.GetDomain(variable)
    if referenceDomain.GetMinimumValue() >= lowerBound:
      return True
    domain = self._domains.get(variable, None)
    if domain==None:
      domain = referenceDomain.Clone()
      self._domains[variable] = domain
    domain.UpdateMinimumValue(lowerBound)
    # sys.stdout.write("UpdateMinimumValue " + str(lowerBound) + " from domain of " + variable.GetName() + "\n")
    return domain.GetSize()>0
  def UpdateMaximumValue(self, variable, upperBound):
    referenceDomain = self.GetDomain(variable)
    if referenceDomain.GetMaximumValue() <= upperBound:
      return True
    domain = self._domains.get(variable, None)
    if domain==None:
      domain = referenceDomain.Clone()
      self._domains[variable] = domain
    domain.UpdateMaximumValue(upperBound)
    # sys.stdout.write("UpdateMaximumValue " + str(upperBound) + " from domain of " + variable.GetName() + "\n")
    return domain.GetSize()>0
  def KeepOnlyValue(self, variable, value):
    domain = self._domains.get(variable, None)
    if domain==None:
      referenceDomain = self.GetDomain(variable)
      domain = referenceDomain.Clone()
      self._domains[variable] = domain
    domain.KeepOnlyValue(value)
    return True

# 3) Solver
# ---------

class SherlockSolver:
  def __init__(self):
    self._falsified = False
    self.numberOfChoices = 0
    self._stack = []
    from SherlockPrinter import SherlockPrinter
    self._printer = SherlockPrinter()

# 3.1) Generate and test
# ----------------------

  def GenerateAndTest(self, problem):
    self._falsified = False
    self.numberOfChoices = 0
    self._stack = []
    assignment = self.BuildInitialAssignment(problem)
    solution = self._GenerateAndTest(problem, assignment)
    if solution==None:
      return []
    return [solution]

  def _GenerateAndTest(self, problem, assignment):
    variable = self.SelectFirstVariable(problem, assignment)
    if variable==None:
      for variable in assignment.GetVariables():
        self.PushVariable(variable)
      satisfiable = self.Propagate(problem, assignment)
      if satisfiable:
        return assignment
      return None
    self.numberOfChoices += 1
    domain = assignment.GetDomain(variable)
    for value in domain.GetValues():
      candidate = assignment.Expand()
      candidate.KeepOnlyValue(variable, value)
      solution = self._GenerateAndTest(problem, candidate)
      if solution!=None:
        return solution
    return None

  def SelectFirstVariable(self, problem, assignment):
    for variable in assignment.GetVariables():
      domain = assignment.GetDomain(variable)
      if domain.GetSize()>1:
        return variable
    return None

# 3.2) Find one solution
# ----------------------

  def FindOneSolution(self, problem):
    self._falsified = False
    self.numberOfChoices = 0
    self._stack = []
    assignment = self.BuildInitialAssignment(problem)
    satisfiable = self.Propagate(problem, assignment)
    # self._printer.WriteAssignment(problem, assignment, sys.stdout)
    if not satisfiable:
      return []
    solution = self._FindOneSolution(problem, assignment)
    if solution==None:
      return []
    return [solution]

  def _FindOneSolution(self, problem, assignment):
    if self.IsSolution(problem, assignment):
      return assignment
    self.numberOfChoices += 1
    variable = self.SelectVariable(problem, assignment)
    domain = assignment.GetDomain(variable)
    for value in domain.GetValues():
      candidate = assignment.Expand()
      candidate.KeepOnlyValue(variable, value)
      self.PushVariable(variable)
      satisfiable = self.Propagate(problem, candidate)
      if satisfiable:
        solution = self._FindOneSolution(problem, candidate)
        if solution!=None:
          return solution
    return None

# 3.3) Find all solutions
# -----------------------

  def FindAllSolutions(self, problem):
    self._falsified = False
    self.numberOfChoices = 0
    self._stack = []
    assignment = self.BuildInitialAssignment(problem)
    satisfiable = self.Propagate(problem, assignment)
    # self._printer.WriteAssignment(problem, assignment, sys.stdout)
    if not satisfiable:
      return []
    solutions = []
    self._FindAllSolutions(problem, assignment, solutions)
    return solutions

  def _FindAllSolutions(self, problem, assignment, solutions):
    if self.IsSolution(problem, assignment):
      solutions.append(assignment)
      return
    self.numberOfChoices += 1
    variable = self.SelectVariable(problem, assignment)
    domain = assignment.GetDomain(variable)
    for value in domain.GetValues():
      candidate = assignment.Expand()
      candidate.KeepOnlyValue(variable, value)
      self.PushVariable(variable)
      satisfiable = self.Propagate(problem, candidate)
      if satisfiable:
        self._FindAllSolutions(problem, candidate, solutions)

# 3.4) Assignment management
# --------------------------

  def BuildInitialAssignment(self, problem):
    assignment = SherlockAssignment()
    for variable in problem.GetVariables():
      domain = variable.GetDomain()
      assignment.AddDomain(variable, domain.Clone())
      self.PushVariable(variable)
    return assignment

  def PushVariable(self, variable):
    if variable.IsStacked():
      return
    self._stack.append(variable)
    variable.SetStacked(True)

  def PopVariable(self):
    if len(self._stack)==0:
      return None
    variable = self._stack.pop()
    variable.SetStacked(False)
    return variable

  def Propagate(self, problem, assignment):
    satisfiable = True
    while len(self._stack)!=0:
      variable = self.PopVariable()
      if satisfiable:
        satisfiable = self.PropagateVariableAssignmentInProblem(problem, assignment, variable)
    return satisfiable

  def PropagateVariableAssignmentInProblem(self, problem, assignment, variable):
    for constraint in variable.GetOccurrences():
      satisfiable = self.PropagateVariableAssignmentInConstraint(problem, constraint, assignment, variable)
      if not satisfiable:
        return False
    return True

  def PropagateVariableAssignmentInConstraint(self, problem, constraint, assignment, variable):
    if constraint.GetType()==SherlockConstraint.EQUAL:
      satisfiable = self.PropagateVariableAssignmentInConstraintEqual(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.DIFFERENT:
      satisfiable = self.PropagateVariableAssignmentInConstraintDifferent(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.LESS:
      satisfiable = self.PropagateVariableAssignmentInConstraintLess(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.LESS_OR_EQUAL:
      satisfiable = self.PropagateVariableAssignmentInConstraintLessOrEqual(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.GREATER:
      satisfiable = self.PropagateVariableAssignmentInConstraintGreater(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.GREATER_OR_EQUAL:
      satisfiable = self.PropagateVariableAssignmentInConstraintGreaterOrEqual(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.ALL_DIFFERENT:
      satisfiable = self.PropagateVariableAssignmentInConstraintAllDifferent(problem, constraint, assignment, variable)
    elif constraint.GetType()==SherlockConstraint.CLAUSE:
      satisfiable = self.PropagateVariableAssignmentInConstraintClause(problem, constraint, assignment, variable)
    else:
      satisfiable = True
    return satisfiable

  def PropagateVariableAssignmentInConstraintEqual(self, problem, constraint, assignment, modifiedVariable):
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      domain = assignment.GetDomain(variable)
      domainSize = domain.GetSize()
      if domainSize==0:
        return False
      coefficient = monomial.GetCoefficient()
      lowerBoundPolynomial = float(self.CalculateLowerBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable))
      upperBoundPolynomial = float(self.CalculateUpperBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable))
      bound = constraint.GetBound()
      if coefficient>0:
        lowerBound = (bound - upperBoundPolynomial) / coefficient
        upperBound = (bound - lowerBoundPolynomial) / coefficient
      else:
        lowerBound = (bound - lowerBoundPolynomial) / coefficient
        upperBound = (bound - upperBoundPolynomial) / coefficient
      ok = assignment.UpdateMinimumValue(variable, lowerBound)
      if not ok:
        return False
      ok = assignment.UpdateMaximumValue(variable, upperBound)
      if not ok:
        return False
      if assignment.GetDomainSize(variable)!=domainSize:
        self.PushVariable(variable)
    return True

  def CalculateLowerBoundOfPolynomial(self, polynomial, assignment, excludedVariable = None):
    lowerBound = 0
    for monomial in polynomial:
      variable = monomial.GetVariable()
      if variable!=excludedVariable:
        coefficient = monomial.GetCoefficient()
        domain = assignment.GetDomain(variable)
        if coefficient<0:
          lowerBound += coefficient*domain.GetMaximumValue()
        else:
          lowerBound += coefficient*domain.GetMinimumValue()
    return lowerBound

  def CalculateUpperBoundOfPolynomial(self, polynomial, assignment, excludedVariable = None):
    upperBound = 0
    for monomial in polynomial:
      variable = monomial.GetVariable()
      if variable!=excludedVariable:
        coefficient = monomial.GetCoefficient()
        domain = assignment.GetDomain(variable)
        if coefficient<0:
          upperBound += coefficient*domain.GetMinimumValue()
        else:
          upperBound += coefficient*domain.GetMaximumValue()
    return upperBound

  def PropagateVariableAssignmentInConstraintDifferent(self, problem, constraint, assignment, modifiedVariable):
    numberOfFreeVariables = 0
    sum = 0
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      coefficient = monomial.GetCoefficient()
      domain = assignment.GetDomain(variable)
      if domain.GetSize()==1:
        value = domain.GetFirstValue()
        sum += coefficient*value
      else:
        numberOfFreeVariables += 1
        if numberOfFreeVariables>1:
          return True
        freeVariable = variable
        freeCoefficient = coefficient
        freeDomain = domain
    if numberOfFreeVariables==0:
      return sum!=constraint.GetBound()
    sum = constraint.GetBound() - sum
    if sum % freeCoefficient != 0:
      return True
    forbiddenValue = sum // freeCoefficient
    if not freeDomain.DoesContainValue(forbiddenValue):
      return True
    ok = assignment.RemoveValue(freeVariable, forbiddenValue)
    if not ok:
      return False
    self.PushVariable(freeVariable)
    return True

  def PropagateVariableAssignmentInConstraintLess(self, problem, constraint, assignment, modifiedVariable):
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      domain = assignment.GetDomain(variable)
      domainSize = domain.GetSize()
      if domainSize==0:
        return False
      coefficient = monomial.GetCoefficient()
      bound = constraint.GetBound()
      lowerBoundPolynomial = self.CalculateLowerBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable)
      if coefficient>0:
        upperBound = (bound - lowerBoundPolynomial) / coefficient
        if (upperBound).is_integer():
          upperBound -= 0.5
        ok = assignment.UpdateMaximumValue(variable, upperBound)
      else:
        lowerBound = (bound - lowerBoundPolynomial) / coefficient
        if (lowerBound).is_integer():
          lowerBound += 0.5
        ok = assignment.UpdateMinimumValue(variable, lowerBound)
      if not ok:
        return False
      if assignment.GetDomainSize(variable)!=domainSize:
        self.PushVariable(variable)
    return True

  def PropagateVariableAssignmentInConstraintLessOrEqual(self, problem, constraint, assignment, modifiedVariable):
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      domain = assignment.GetDomain(variable)
      domainSize = domain.GetSize()
      if domainSize==0:
        return False
      coefficient = monomial.GetCoefficient()
      bound = constraint.GetBound()
      lowerBoundPolynomial = self.CalculateLowerBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable)
      if coefficient>0:
        upperBound = (bound - lowerBoundPolynomial) / coefficient
        ok = assignment.UpdateMaximumValue(variable, upperBound)
      else:
        lowerBound = (bound - lowerBoundPolynomial) / coefficient
        ok = assignment.UpdateMinimumValue(variable, lowerBound)
      if not ok:
        return False
      if assignment.GetDomainSize(variable)!=domainSize:
        self.PushVariable(variable)
    return True

  def PropagateVariableAssignmentInConstraintGreater(self, problem, constraint, assignment, modifiedVariable):
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      domain = assignment.GetDomain(variable)
      domainSize = domain.GetSize()
      if domainSize==0:
        return False
      coefficient = monomial.GetCoefficient()
      bound = constraint.GetBound()
      upperBoundPolynomial = self.CalculateUpperBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable)
      if coefficient>0:
        lowerBound = (bound - upperBoundPolynomial) / coefficient
        if (lowerBound).is_integer():
          lowerBound += 0.5
        ok = assignment.UpdateMinimumValue(variable, lowerBound)
      else:
        upperBound = (bound - upperBoundPolynomial) / coefficient
        if (upperBound).is_integer():
          upperBound -= 0.5
        ok = assignment.UpdateMaximumValue(variable, upperBound)
      if not ok:
        return False
      if assignment.GetDomainSize(variable)!=domainSize:
        self.PushVariable(variable)
    return True

  def PropagateVariableAssignmentInConstraintGreaterOrEqual(self, problem, constraint, assignment, modifiedVariable):
    for monomial in constraint.GetPolynomial():
      variable = monomial.GetVariable()
      domain = assignment.GetDomain(variable)
      domainSize = domain.GetSize()
      if domainSize==0:
        return False
      coefficient = monomial.GetCoefficient()
      bound = constraint.GetBound()
      upperBoundPolynomial = self.CalculateUpperBoundOfPolynomial(constraint.GetPolynomial(), assignment, variable)
      if coefficient>0:
        lowerBound = (bound - upperBoundPolynomial) / coefficient
        ok = assignment.UpdateMinimumValue(variable, lowerBound)
      else:
        upperBound = (bound - upperBoundPolynomial) / coefficient
        ok = assignment.UpdateMaximumValue(variable, upperBound)
      if not ok:
        return False
      if assignment.GetDomainSize(variable)!=domainSize:
        self.PushVariable(variable)
    return True

  def PropagateVariableAssignmentInConstraintAllDifferent(self, problem, constraint, assignment, variable):
    domain = assignment.GetDomain(variable)
    if domain.GetSize()>1:
      return True
    value = domain.GetFirstValue()
    for var in constraint.GetVariables():
      if var!=variable:
        dom = assignment.GetDomain(var)
        if dom.DoesContainValue(value):
          ok = assignment.RemoveValue(var, value)
          if not ok:
            return False
          self.PushVariable(var)
    return True

  def PropagateVariableAssignmentInConstraintClause(self, problem, constraint, assignment, variable):
    freeLiteral = None
    for literal in constraint.GetLiterals():
      value = self.AssessLiteralValue(literal, assignment)
      if value==1:
        return True
      elif value==0:
        if freeLiteral==None:
          freeLiteral = literal
        else:
          return True
    if freeLiteral==None:
      return False
    self.KeepSatisfyingValuesOnly(freeLiteral, assignment)
    return True

  def AssessLiteralValue(self, literal, assignment):
    variable = literal.GetVariable()
    domain = assignment.GetDomain(variable)
    value = literal.GetValue()
    acceptableValueCount = 0
    for candidateValue in domain.GetValues():
      if literal.GetType()==SherlockLiteral.EQUAL:
        if candidateValue==value:
          acceptableValueCount += 1
      elif literal.GetType()==SherlockLiteral.DIFFERENT:
        if candidateValue!=value:
          acceptableValueCount += 1
      elif literal.GetType()==SherlockLiteral.LESS:
        if candidateValue<value:
          acceptableValueCount += 1
      elif literal.GetType()==SherlockLiteral.LESS_OR_EQUAL:
        if candidateValue<=value:
          acceptableValueCount += 1
      elif literal.GetType()==SherlockLiteral.GREATER:
        if candidateValue>value:
          acceptableValueCount += 1
      elif literal.GetType()==SherlockLiteral.GREATER_OR_EQUAL:
        if candidateValue>=value:
          acceptableValueCount += 1
    if acceptableValueCount==0:
      return -1
    # if acceptableValueCount==1 and domain.GetSize()==1:
      # return 1
    return 0

  def KeepSatisfyingValuesOnly(self, literal, assignment):
    variable = literal.GetVariable()
    domain = assignment.GetDomain(variable)
    size = domain.GetSize()
    value = literal.GetValue()
    selectedValue = None
    for candidateValue in domain.GetValues():
      if literal.GetType()==SherlockLiteral.EQUAL:
        if candidateValue!=value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
      elif literal.GetType()==SherlockLiteral.DIFFERENT:
        if candidateValue==value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
      elif literal.GetType()==SherlockLiteral.LESS:
        if candidateValue>=value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
      elif literal.GetType()==SherlockLiteral.LESS_OR_EQUAL:
        if candidateValue>value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
      elif literal.GetType()==SherlockLiteral.GREATER:
        if candidateValue<=value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
      elif literal.GetType()==SherlockLiteral.GREATER_OR_EQUAL:
        if candidateValue<value:
          ok = assignment.RemoveValue(variable, candidateValue)
          if not ok:
            return False
    domain = assignment.GetDomain(variable)
    if domain.GetSize()!=size:
      self.PushVariable(variable)
    return True

  def IsSolution(self, problem, assignment):
    for variable in problem.GetVariables():
      if assignment.GetDomainSize(variable)>1:
        return False
    return True

  def SelectVariable(self, problem, assignment):
    return self.SelectSmallestDomainVariable(problem, assignment)

  def SelectSmallestDomainVariable(self, problem, assignment):
    smallestSize = 100000
    selectedVariable = None
    for variable in assignment.GetVariables():
      domain = assignment.GetDomain(variable)
      size = domain.GetSize()
      if size>1 and size<smallestSize:
        smallestSize = size
        selectedVariable = variable
    return selectedVariable

# 3.5) Find best solution
# -----------------------

  def FindBestSolution(self, problem):
    self.numberOfChoices = 0
    self._falsified = False
    if problem.GetObjectivePolynomial()==None:
      return self.FindOneSolution(problem)
    assignment = self.BuildInitialAssignment(problem)
    satisfiable = self.Propagate(problem, assignment)
    if not satisfiable:
      return []
    self._bestSolution = None
    if problem.IsMinimizationProblem():
      self._bestScore = self.CalculateUpperBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment) + 1
    else:
      self._bestScore = self.CalculateLowerBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment) - 1
    self._stack = []
    self._FindBestSolution(problem, assignment)
    if self._bestSolution==None:
      return []
    return [self._bestSolution]

  def _FindBestSolution(self, problem, assignment):
    if self.IsSolution(problem, assignment):
      if problem.IsMinimizationProblem():
        score = self.CalculateLowerBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
        if score<self._bestScore:
          self._bestScore = score
          self._bestSolution = assignment
          # sys.stdout.write("new best score: %d\n" % self._bestScore)
      else:
        score = self.CalculateUpperBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
        if score>self._bestScore:
          self._bestScore = score
          self._bestSolution = assignment
          # sys.stdout.write("new best score: %d\n" % self._bestScore)
      return
    self.numberOfChoices += 1
    variable = self.SelectVariable(problem, assignment)
    domain = assignment.GetDomain(variable)
    for value in domain.GetValues():
      candidate = assignment.Expand()
      candidate.KeepOnlyValue(variable, value)
      self.PushVariable(variable)
      satisfiable = self.Propagate(problem, candidate)
      if satisfiable:
        self._FindBestSolution(problem, candidate)


# 3.6) Branch and bound
# ---------------------

  def BranchAndBound(self, problem):
    self.numberOfChoices = 0
    self._falsified = False
    if problem.GetObjectivePolynomial()==None:
      return self.FindOneSolution(problem)
    assignment = self.BuildInitialAssignment(problem)
    satisfiable = self.Propagate(problem, assignment)
    # self._printer.WriteAssignment(problem, assignment, sys.stdout)
    if not satisfiable:
      return []
    self._bestSolution = None
    if problem.IsMinimizationProblem():
      self._bestScore = self.CalculateUpperBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment) + 1
    else:
      self._bestScore = self.CalculateLowerBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment) - 1
    self._stack = []
    self._BranchAndBound(problem, assignment)
    if self._bestSolution==None:
      return []
    return [self._bestSolution]

  def _BranchAndBound(self, problem, assignment):
    if self.IsSolution(problem, assignment):
      if problem.IsMinimizationProblem():
        score = self.CalculateLowerBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
        if score<self._bestScore:
          self._bestScore = score
          self._bestSolution = assignment
          # sys.stdout.write("new best score: %d\n" % self._bestScore)
      else:
        score = self.CalculateUpperBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
        if score>self._bestScore:
          self._bestScore = score
          self._bestSolution = assignment
          # sys.stdout.write("new best score: %d\n" % self._bestScore)
      return
    if problem.IsMinimizationProblem():
      bestPossibleScore = self.CalculateLowerBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
      if bestPossibleScore>=self._bestScore:
        return
    else:
      bestPossibleScore = self.CalculateUpperBoundOfPolynomial(problem.GetObjectivePolynomial(), assignment)
      if bestPossibleScore<=self._bestScore:
        return
    self.numberOfChoices += 1
    variable = self.SelectVariable(problem, assignment)
    domain = assignment.GetDomain(variable)
    for value in domain.GetValues():
      candidate = assignment.Expand()
      candidate.KeepOnlyValue(variable, value)
      self.PushVariable(variable)
      satisfiable = self.Propagate(problem, candidate)
      if satisfiable:
        self._BranchAndBound(problem, candidate)

